
import asyncio

import xml.etree.ElementTree as ET

from datetime import datetime, timezone

import logging
logger = logging.getLogger(__name__)

# All xml data received from the remote connection should be contained in one of the following tags
TAGS = (b'message',
        b'newTextVector',
        b'newNumberVector',
        b'newSwitchVector',
        b'newBLOBVector',
        b'delProperty',
        b'defSwitchVector',
        b'setSwitchVector',
        b'defLightVector',
        b'setLightVector',
        b'defTextVector',
        b'setTextVector',
        b'defNumberVector',
        b'setNumberVector',
        b'defBLOBVector',
        b'setBLOBVector',
        b'getProperties'
       )


# Note these are strings, as they are used for checking xmldata.tag values

DEFTAGS = ( 'defSwitchVector',
            'defLightVector',
            'defTextVector',
            'defNumberVector',
            'defBLOBVector'
          )

NEWTAGS = ('newTextVector',
           'newNumberVector',
           'newSwitchVector',
           'newBLOBVector'
          )


# _STARTTAGS is a tuple of ( b'<defTextVector', ...  ) data received will be tested to start with such a starttag
_STARTTAGS = tuple(b'<' + tag for tag in TAGS)

# _ENDTAGS is a tuple of ( b'</defTextVector>', ...  ) data received will be tested to end with such an endtag
_ENDTAGS = tuple(b'</' + tag + b'>' for tag in TAGS)


def _makestart(element):
    "Given an xml element, returns a string of its start, including < tag attributes >"
    attriblist = ["<", element.tag]
    for key,value in element.attrib.items():
        attriblist.append(f" {key}=\"{value}\"")
    attriblist.append(">")
    return "".join(attriblist)


class RemoteConnection:

    def __init__(self, host, port,
                       blob_enable,
                       debug_enable,
                       alldrivers,      # list of drivers connected to the server
                       remotes,         # list of RemoteConnection(s) connected to the server
                       serverwriterque, # queue of items for the server to transmit
                       connectionpool   # pool of connections calling this server
                 ):

        self.indihost = host
        self.indiport = port

        #  These can be True or False.
        self.blob_enable = blob_enable
        self.debug_enable = debug_enable

        # This is a set of devicenames learnt on this remote connection
        # populated as devices are learnt, and used to check for duplicate names
        self.devicenames = set()

        self.alldrivers = alldrivers
        self.remotes= remotes
        self.serverwriterque = serverwriterque
        self.connectionpool = connectionpool

        # create queue where client will put xml data to be transmitted
        self._writerque = asyncio.Queue(4)

        # and create readerque where received xmldata will be put
        self._readerque = asyncio.Queue(4)

        # self.connected is True if connection has been made
        self.connected = False

        # and shutdown routine sets this to True to stop coroutines
        self._stop = False
        # this is set when asyncrun is finished
        self.stopped = asyncio.Event()


    async def _hardware(self):
        """Flag connection made or failed messages into serverwriterque"""
        # create a flag so 'remote connection made' message is only set once
        isconnected = False
        while not self._stop:
            await asyncio.sleep(0.1)
            if self.connected:
                if isconnected:
                    continue
                isconnected = True
                # a new connection has been made
                await self.send_getProperties()
                for clientconnection in self.connectionpool:
                    if clientconnection.connected:
                        # a client is connected, send a message
                        timestamp = datetime.now(tz=timezone.utc)
                        timestamp = timestamp.replace(tzinfo = None)
                        tstring = timestamp.isoformat(sep='T')
                        messagedata = ET.Element('message')
                        messagedata.set("timestamp", tstring)
                        messagedata.set("message", f"Remote connection made to {self.indihost}:{self.indiport}")
                        await self.queueput(self.serverwriterque, messagedata)
                        break
            else:
                # The connection has failed
                if isconnected:
                    isconnected = False
                    self.devicenames.clear()
                    for clientconnection in self.connectionpool:
                        if clientconnection.connected:
                            # a client is connected, send a message
                            timestamp = datetime.now(tz=timezone.utc)
                            timestamp = timestamp.replace(tzinfo = None)
                            tstring = timestamp.isoformat(sep='T')
                            messagedata = ET.Element('message')
                            messagedata.set("timestamp", tstring)
                            messagedata.set("message", f"Remote connection to {self.indihost}:{self.indiport} lost")
                            await self.queueput(self.serverwriterque, messagedata)
                            break


    def shutdown(self):
        "Shuts down the client, sets the flag self._stop to True"
        self._stop = True

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the client"
        return self._stop


    async def queueput(self, queue, value, timeout=0.5):
        """Method used internally, but available if usefull.
           Given an asyncio.Queue object attempts to put value into the queue.
           If the queue is full, checks self._stop every timeout seconds.
           Returns True when the value is added to queue,
           or False if self._stop is True and the value not added."""
        while not self._stop:
            try:
                await asyncio.wait_for(queue.put(value), timeout)
            except asyncio.TimeoutError:
                # queue is full, continue while loop, checking stop flag
                continue
            return True
        return False


    async def warning(self, message):
        """The given string message will be logged at level WARNING,
           and will be injected into the received data, which will be
           picked up by the rxevent method.
           It is a way to set a message on to your client display, in the
           same way messages come from the INDI service."""
        try:
            logger.warning(message)
            timestamp = datetime.now(tz=timezone.utc)
            timestamp = timestamp.replace(tzinfo=None)
            root = ET.fromstring(f"<message timestamp=\"{timestamp.isoformat(sep='T')}\" message=\"{message}\" />")
            # and place root into readerque
            await self.queueput(self._readerque, root)
        except Exception :
            logger.exception("Exception report from RemoteConnection.warning method")


    async def _comms(self):
        "Create a connection to an INDI port"
        try:
            while not self._stop:
                t1 = None
                t2 = None
                try:
                    # start by openning a connection
                    await self.warning(f"Attempting to connect to {self.indihost}:{self.indiport}")
                    reader, writer = await asyncio.open_connection(self.indihost, self.indiport)
                    self.connected = True
                    await self.warning(f"Connected to {self.indihost}:{self.indiport}")
                    t1 = asyncio.create_task(self._run_tx(writer))
                    t2 = asyncio.create_task(self._run_rx(reader))
                    await asyncio.gather(t1, t2)
                except ConnectionRefusedError:
                    await self.warning(f"Connection refused on {self.indihost}:{self.indiport}")
                except ConnectionError:
                    await self.warning(f"Connection Lost on {self.indihost}:{self.indiport}")
                except OSError:
                    await self.warning(f"Connection Error on {self.indihost}:{self.indiport}")
                except Exception:
                    logger.exception(f"Connection Error on {self.indihost}:{self.indiport}")
                    await self.warning("Connection failed")
                self.connected = False
                # connection has failed, ensure all tasks are done
                if t1:
                    while not t1.done():
                        await asyncio.sleep(0)
                if t2:
                    while not t2.done():
                        await asyncio.sleep(0)
                if self._stop:
                    break
                else:
                    await self.warning("Connection failed, re-trying...")
                # wait five seconds before re-trying, but keep checking
                # that self._stop has not been set
                count = 0
                while not self._stop:
                    await asyncio.sleep(0.5)
                    count += 1
                    if count >= 10:
                        break
        except Exception:
            logger.exception("Exception report from RemoteConnection._comms method")
            raise
        finally:
            self.shutdown()


    async def send(self, xmldata):
        """Transmits xmldata, which is an xml.etree.ElementTree object"""
        if self.connected and (not self._stop):
            await self.queueput(self._writerque, xmldata)


    def _logtx(self, txdata):
        "log tx data with level debug"
        if not self.debug_enable:
            return
        binarydata = ET.tostring(txdata)
        logger.debug("TX:: " + binarydata.decode())


    async def _run_tx(self, writer):
        "Monitors self._writerque and if it has data, uses writer to send it"
        try:
            while self.connected and (not self._stop):
                try:
                    txdata = await asyncio.wait_for(self._writerque.get(), timeout=0.5)
                except asyncio.TimeoutError:
                    continue
                self._writerque.task_done()
                if not self.connected:
                    break
                if self._stop:
                    break
                if not self.blob_enable:
                    if (txdata.tag == "setBLOBVector") or  (txdata.tag == "newBLOBVector"):
                        # blobs not enabled
                        continue
                # send it out on the port
                binarydata = ET.tostring(txdata)
                writer.write(binarydata)
                await writer.drain()
                # data has been transmitted
                if logger.isEnabledFor(logging.DEBUG):
                    self._logtx(txdata)
        except ConnectionError:
            raise
        except Exception:
            logger.exception("Exception report from RemoteConnection._run_tx method")
            raise
        finally:
            writer.close()
            await writer.wait_closed()
            self.connected = False

    def _logrx(self, rxdata):
        "log rx data"
        if not self.debug_enable:
            return
        binarydata = ET.tostring(rxdata)
        logger.debug("RX:: " + binarydata.decode())


    async def _run_rx(self, reader):
        "pass xml.etree.ElementTree data to readerque"
        try:
            # get block of xml.etree.ElementTree data
            # from self._xmlinput and append it to  readerque
            while self.connected and (not self._stop):
                rxdata = await self._xmlinput(reader)
                if rxdata is None:
                    return
                # and place rxdata into readerque
                result = await self.queueput(self._readerque, rxdata, 0.2)
                if not result:
                    # self._stop must be set
                    return
                # rxdata in readerque, log it, then continue with next block
                if logger.isEnabledFor(logging.DEBUG):
                    self._logrx(rxdata)
        except ConnectionError:
            raise
        except Exception:
            logger.exception("Exception report from RemoteConnection._run_rx")
            raise


    async def _xmlinput(self, reader):
        """get received data, parse it, and return it as xml.etree.ElementTree object
           Returns None if notconnected/stop flags arises"""
        message = b''
        messagetagnumber = None
        while self.connected and (not self._stop):
            await asyncio.sleep(0)
            data = await self._datainput(reader)
            # data is either None, or binary data ending in b">"
            if data is None:
                return
            if not self.connected:
                return
            if self._stop:
                return
            if not message:
                # data is expected to start with <tag, first strip any newlines
                data = data.strip()
                for index, st in enumerate(_STARTTAGS):
                    if data.startswith(st):
                        messagetagnumber = index
                        break
                    elif st in data:
                        # remove any data prior to a starttag
                        positionofst = data.index(st)
                        data = data[positionofst:]
                        messagetagnumber = index
                        break
                else:
                    # data does not start with a recognised tag, so ignore it
                    # and continue waiting for a valid message start
                    continue
                # set this data into the received message
                message = data
                # either further children of this tag are coming, or maybe its a single tag ending in "/>"
                if message.endswith(b'/>'):
                    # the message is complete, handle message here
                    try:
                        root = ET.fromstring(message.decode("us-ascii"))
                    except ET.ParseError:
                       # failed to parse the message, continue at beginning
                        message = b''
                        messagetagnumber = None
                        continue
                    # xml datablock done, return it
                    return root
                # and read either the next message, or the children of this tag
                continue
            # To reach this point, the message is in progress, with a messagetagnumber set
            # keep adding the received data to message, until an endtag is reached
            message += data
            if message.endswith(_ENDTAGS[messagetagnumber]):
                # the message is complete, handle message here
                try:
                    root = ET.fromstring(message.decode("us-ascii"))
                except ET.ParseError:
                    # failed to parse the message, continue at beginning
                    message = b''
                    messagetagnumber = None
                    continue
                # xml datablock done, return it
                return root
            # so message is in progress, with a messagetagnumber set
            # but no valid endtag received yet, so continue the loop


    async def _datainput(self, reader):
        """Waits for binary string of data ending in > from the port
           Returns None if notconnected/stop flags arises"""
        binarydata = b""
        while self.connected and (not self._stop):
            await asyncio.sleep(0)
            try:
                data = await reader.readuntil(separator=b'>')
            except asyncio.LimitOverrunError:
                data = await reader.read(n=32000)
            except asyncio.IncompleteReadError:
                binarydata = b""
                await asyncio.sleep(0.1)
                continue
            if not data:
                await asyncio.sleep(0.01)
                continue
            # data received
            if b">" in data:
                binarydata = binarydata + data
                return binarydata
            # data has content but no > found
            binarydata += data
            # could put a max value here to stop this increasing indefinetly


    async def _rxhandler(self):
        """Handles received data from self._readerque"""
        try:
            while not self._stop:
                # get block of data from the self._readerque
                try:
                    rxdata = await asyncio.wait_for(self._readerque.get(), timeout=0.5)
                except asyncio.TimeoutError:
                    # nothing to read, continue while loop which re-checks the _stop flag
                    continue
                self._readerque.task_done()

                devicename = rxdata.get("device")
                vectorname = rxdata.get("name")

                if not self.blob_enable:
                    if rxdata.tag in ("setBLOBVector", "newBLOBVector"):
                        # blobs not enabled
                        continue

                if devicename:
                    if rxdata.tag in DEFTAGS:
                        # check for duplicate devicename
                        for driver in self.alldrivers:
                            if devicename in driver:
                                logger.error(f"A duplicate devicename {devicename} has been detected")
                                await self.queueput(self.serverwriterque, None)
                                return
                        for remote in self.remotes:
                            if remote is self:
                                continue
                            if devicename in remote.devicenames:
                                logger.error(f"A duplicate devicename {devicename} has been detected")
                                await self.queueput(self.serverwriterque, None)
                                return
                        if devicename not in self.devicenames:
                            self.devicenames.add(devicename)
                        if rxdata.tag == "defBLOBVector":
                            # every time a defBLOBVector is received, send an enable BLOB instruction
                            xmldata = ET.Element('enableBLOB')
                            xmldata.set("device", devicename)
                            xmldata.set("name", vectorname)
                            if self.blob_enable:
                                xmldata.text = "Also"
                            else:
                                xmldata.text = "Never"
                            await self.send(xmldata)

                    # if a new vector or a getProperties has been received, and is targetted at a
                    # driver, send it to the driver and nowhere else
                    if (rxdata.tag in NEWTAGS) or (rxdata.tag == "getProperties"):
                        dfound = False
                        for driver in self.alldrivers:
                            if devicename in driver:
                                await self.queueput(driver.readerque, rxdata)
                                # no need to transmit this anywhere else
                                dfound = True
                                break
                        if dfound:
                            continue

                # so not targetted at a local known devicename
                # transmit to drivers if rxdata is either a getProperties or because the driver is snooping on it

                if rxdata.tag == "getProperties":
                    for driver in self.alldrivers:
                        # either no devicename, or an unknown device, so send to all drivers
                        await self.queueput(driver.readerque, rxdata)
                elif rxdata.tag not in NEWTAGS:
                    for driver in self.alldrivers:
                        # Check if this driver is snooping on this device/vector
                        if driver.snoopall:
                            await self.queueput(driver.readerque, rxdata)
                        elif devicename and (devicename in driver.snoopdevices):
                            await self.queueput(driver.readerque, rxdata)
                        elif devicename and vectorname and ((devicename, vectorname) in driver.snoopvectors):
                            await self.queueput(driver.readerque, rxdata)

                # transmit rxdata out to clients

                # If no clients are connected, do not put this data into
                # the serverwriterque
                for clientconnection in self.connectionpool:
                    if clientconnection.connected:
                        # at least one is connected, so this data is put into
                        # serverwriterque
                        await self.queueput(self.serverwriterque, rxdata)
                        break

                # transmit rxdata out to other remote connections
                for remcon in self.remotes:
                    if remcon is self:
                        continue
                    await remcon.send(rxdata)

        except Exception:
            logger.exception("Exception report from RemoteConnection._rxhandler method")
            raise
        finally:
            self.shutdown()


    async def send_getProperties(self):
        """Sends a getProperties request"""
        if self.connected:
            xmldata = ET.Element('getProperties')
            xmldata.set("version", "1.7")
            await self.send(xmldata)


    async def asyncrun(self):
        "Await this method to run the connectiont."
        self._stop = False
        try:
            await asyncio.gather(self._comms(), self._rxhandler(), self._hardware())
        except asyncio.CancelledError:
            self._stop = True
            raise
        finally:
            self.stopped.set()
            self._stop = True





import asyncio, logging

from datetime import datetime, timezone

import xml.etree.ElementTree as ET

from .ipydriver import IPyDriver

from .remote import RemoteConnection

from .exdriver import ExDriver

logger = logging.getLogger(__name__)


# All xml data should be contained in one of the following tags
TAGS = (b'getProperties',
        b'newTextVector',
        b'newNumberVector',
        b'newSwitchVector',
        b'newBLOBVector',
        b'enableBLOB',
        b'message',
        b'delProperty',
        b'defSwitchVector',
        b'setSwitchVector',
        b'defLightVector',
        b'setLightVector',
        b'defTextVector',
        b'setTextVector',
        b'defNumberVector',
        b'setNumberVector',
        b'defBLOBVector',
        b'setBLOBVector'
       )


# Note these are strings, as they are used for checking xmldata.tag values

NEWTAGS = ('newTextVector',
           'newNumberVector',
           'newSwitchVector',
           'newBLOBVector'
          )


# _STARTTAGS is a tuple of ( b'<defTextVector', ...  ) data received will be tested to start with such a starttag
_STARTTAGS = tuple(b'<' + tag for tag in TAGS)


# _ENDTAGS is a tuple of ( b'</defTextVector>', ...  ) data received will be tested to end with such an endtag
_ENDTAGS = tuple(b'</' + tag + b'>' for tag in TAGS)



def cleanque(que):
    "Clears out a que"
    try:
        while True:
            que.get_nowait()
            que.task_done()
    except asyncio.QueueEmpty:
        # que is now empty, nothing further to do
        pass


class IPyServer:

    """Once an instance of this class is created, the asyncrun method
       should be awaited which will open a port, and the INDI service
       will be available for clients to connect.

       drivers are IPyDriver objects this driver handles,
       host and port are "localhost" and 7624 as default

       maxconnections is the number of simultaneous client connections
       accepted, with a default of 5. The number given should be
       between 1 and 10 inclusive.

       If, prior to asyncrun being awaited, the add_remote method is called,
       then a connection will be made to a remote INDI server and any of its
       drivers.

       The add_remote method can be called multiple times to create
       connections to different servers making a branching
       tree of servers and drivers.

       The add_exdriver method can be called to run an executable driver and
       this server will communicate to it by stdin, stdout and stderr, therefore
       ipyserver can act as a general INDI server for third party drivers as
       well as ipydriver instances.
       """


    def __init__(self, *drivers, host="localhost", port=7624, maxconnections=5):

        self.drivers = list(drivers)
        self.host = host
        self.port = port

        # traffic is transmitted out on the serverwriterque
        self.serverwriterque = asyncio.Queue(6)
        # and read in from the serverreaderque
        self.serverreaderque = asyncio.Queue(6)

        # If True, xmldata will be logged at DEBUG level
        self.debug_enable = True

        if maxconnections<1 or maxconnections>10:
            raise ValueError("maxconnections should be a number between 1 and 10 inclusive.")
        self.maxconnections = maxconnections

        # this is a dictionary of device name to device
        self.devices = {}

        # self.remotes is a list of RemoteConnection objects running connections to remote servers
        # this list is populated by calling self.add_remote(host, port, debug_enable)
        self.remotes = []

        # self.exdrivers is a list of ExDriver objects running external drivers
        # this list is populated by calling self.add_exdriver(program, *args, debug_enable=False)
        self.exdrivers = []

        for driver in self.drivers:
            if not isinstance(driver, IPyDriver):
                raise TypeError("The drivers set in IPyServer must all be IPyDrivers")
            if driver.comms is not None:
                 raise RuntimeError("A driver communications method has already been set, there can only be one")
            for devicename in driver:
                if devicename in self.devices:
                    # duplicate devicename
                    raise ValueError(f"Device name {devicename} is duplicated in the attached drivers.")
            self.devices.update(driver.data)

        self.connectionpool = []
        for connection_id in range(0, maxconnections):
            self.connectionpool.append(_ClientConnection(connection_id, self.serverreaderque))

        # This alldrivers list will have exdrivers added to it, so the list
        # here is initially a copy of self.drivers
        self.alldrivers = self.drivers.copy()

        for driver in self.drivers:
            # an instance of _DriverComms is created for each driver
            # each _DriverComms object has lists of drivers and remotes
            # these will be used to send snooping traffic

            driver.comms = _DriverComms(driver,
                                        self.serverwriterque,
                                        self.connectionpool,
                                        self.alldrivers,
                                        self.remotes)
        # shutdown routine sets this to True to stop coroutines
        self._stop = False
        # this is set when asyncrun is finished
        self.stopped = asyncio.Event()
        self.server = None

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the server"
        return self._stop

    def shutdown(self, shutdownmessage=""):
        """Shuts down the server, sets the flag self._stop to True
           and sends shutdownmessage to logger.error if given"""
        if shutdownmessage:
            logger.error(shutdownmessage)
        self._stop = True
        for driver in self.drivers:
            driver.shutdown()
        for remcon in self.remotes:
            remcon.shutdown()
        for exd in self.exdrivers:
            exd.shutdown()
        for clientconnection in self.connectionpool:
            clientconnection.shutdown()
        if self.server is not None:
            self.server.close()

    async def _queueput(self, queue, value, timeout=0.5):
        while not self._stop:
            try:
                await asyncio.wait_for(queue.put(value), timeout)
            except asyncio.TimeoutError:
                # queue is full, continue while loop, checking stop flag
                continue
            break

    def add_remote(self, host, port, blob_enable=False, debug_enable=False):
        """Adds a connection to a remote server.
           blob_enable can be True or False.
           If True BLOBs and other vectors can all be sent.
           If False, then BLOB traffic will not pass over this link.

           If debug_enable is True, then DEBUG level logging will record xml
           traffic, if False, the xml traffic will not be logged. This can be
           used to prevent multiple such connections all logging xml traffic together."""


        remcon = RemoteConnection( host=host, port=port,
                                   blob_enable = blob_enable,
                                   debug_enable = debug_enable,
                                   alldrivers = self.alldrivers,
                                   remotes = self.remotes,
                                   serverwriterque = self.serverwriterque,
                                   connectionpool = self.connectionpool )

        # store this object
        self.remotes.append(remcon)


    def add_exdriver(self, program, *args, debug_enable=False):
        """Adds an executable driver program, runs it and communicates to it via stdin, stdout
           and stderr. Then serves the driver, and any others added, by the listening port.
           args is used for the program arguments if any.
           Any program output on stderr will be logged at level ERROR.

           If debug_enable is True, then DEBUG level logging will record xml
           traffic, if False, the xml traffic will not be logged. This can be
           used to prevent multiple drivers all logging xml traffic together."""
        exd = ExDriver(program, *args, debug_enable=debug_enable)
        # add this exdriver to alldrivers
        self.alldrivers.append(exd)
        # Create a DriverComms object
        exd.comms = _DriverComms(exd,
                                 self.serverwriterque,
                                 self.connectionpool,
                                 self.alldrivers,
                                 self.remotes)
        # store this object
        self.exdrivers.append(exd)



    async def _runserver(self):
        "Runs the server on the given host and port"
        logger.info(f"{self.__class__.__name__} listening on {self.host} : {self.port}")
        self.server = await asyncio.start_server(self.handle_data, self.host, self.port)
        try:
            async with self.server:
                await self.server.serve_forever()
        finally:
            self.shutdown()

    async def handle_data(self, reader, writer):
        "Used by asyncio.start_server, called to handle a client connection"

        for clientconnection in self.connectionpool:
            if not clientconnection.connected:
                # this clientconnection is available
                await clientconnection.handle_data(reader, writer)
                break
        else:
            # no clientconnection is available
            writer.close()
            await writer.wait_closed()


    async def asyncrun(self):
        """await this to operate the server together with its
           drivers and any remote connections."""

        # Note this gather has return_exceptions=True, so exceptions do not stop or cancel
        # other tasks from running, or stop the running loop.
        # This requires each coroutine in this gather to call shutdown on the server if any exception
        # occurs in the coroutine

        self._stop = False
        driverruns = [ self._runners(driver) for driver in self.drivers ]
        remoteruns = [ self._runners(remoteconnection) for remoteconnection in self.remotes ]
        externalruns = [ self._runners(exd) for exd in self.exdrivers ]
        try:
            await asyncio.gather(*driverruns,
                                 *remoteruns,
                                 *externalruns,
                                 self._runserver(),
                                 self._copyfromserver(),
                                 self._sendtoclient(),
                                 return_exceptions=True
                                 )
        finally:
            self.stopped.set()
            self._stop = True


    async def _runners(self, itemtorun):
        """Used by above to run drivers, remotes and externals
           If any shuts down, then calls shutdown the server"""
        try:
            await itemtorun.asyncrun()
        finally:
            self.shutdown()


    async def _copyfromserver(self):
        """Gets data from serverreaderque.
           For every driver, copy data, if applicable, to driver.readerque
           And for every remote connection if applicable, to its send method"""
        try:
            while not self._stop:
                try:
                    quedata = await asyncio.wait_for(self.serverreaderque.get(), 0.5)
                except asyncio.TimeoutError:
                    continue
                self.serverreaderque.task_done()
                connection_id, xmldata = quedata
                devicename = xmldata.get("device")
                propertyname = xmldata.get("name")

                if logger.isEnabledFor(logging.DEBUG) and self.debug_enable:
                    binarydata = ET.tostring(xmldata)
                    logger.debug(f"RX:: {binarydata.decode('utf-8')}")

                exdriverfound = False
                if (xmldata.tag in NEWTAGS) or (xmldata.tag == "getProperties"):
                    # if targetted at a known device, send it to that device
                    if devicename:
                        if devicename in self.devices:
                            # this new or getProperties request is meant for an attached device
                            await self._queueput(self.devices[devicename].driver.readerque, xmldata)
                            # no need to transmit this anywhere else, continue the while loop
                            continue
                        for exd in self.exdrivers:
                            if devicename in exd:
                                # this getProperties request is meant for an external driver
                                await self._queueput(exd.readerque, xmldata)
                                exdriverfound = True
                                break

                if exdriverfound:
                    # no need to transmit this anywhere else, continue the while loop
                    continue

                # copy to all server connections, apart from the one it came in on
                for clientconnection in self.connectionpool:
                    if clientconnection.connected and clientconnection.connection_id != connection_id:
                        await self._queueput(clientconnection.txque, xmldata)

                # copy to all remote connections
                if xmldata.tag != "enableBLOB":
                    for remcon in self.remotes:
                        if not remcon.connected:
                            continue
                        await remcon.send(xmldata)


                # transmit xmldata out to exdrivers,
                if xmldata.tag != "enableBLOB":
                    # enableBLOB instructions are not forwarded to external drivers
                    for driver in self.exdrivers:
                        if xmldata.tag == "getProperties":
                            # either no devicename, or an unknown device
                            await self._queueput(driver.readerque, xmldata)
                        elif xmldata.tag not in NEWTAGS:
                            # either devicename is unknown, or this data is to/from another driver.
                            # So check if this driver is snooping on this device/vector
                            # only forward def's and set's, not 'new' vectors which
                            # do not come from a device, but only from a client to the target device.
                            if driver.snoopall:
                                await self._queueput(driver.readerque, xmldata)
                            elif devicename and (devicename in driver.snoopdevices):
                                await self._queueput(driver.readerque, xmldata)
                            elif devicename and propertyname and ((devicename, propertyname) in driver.snoopvectors):
                                await self._queueput(driver.readerque, xmldata)


                # transmit xmldata out to drivers
                for driver in self.drivers:
                    if xmldata.tag == "getProperties":
                        # either no devicename, or an unknown device
                        await self._queueput(driver.readerque, xmldata)
                    elif xmldata.tag not in NEWTAGS:
                        # either devicename is unknown, or this data is to/from another driver.
                        # So check if this driver is snooping on this device/vector
                        # only forward def's and set's, not 'new' vectors which
                        # do not come from a device, but only from a client to the target device.
                        if driver.snoopall:
                            await self._queueput(driver.readerque, xmldata)
                        elif devicename and (devicename in driver.snoopdevices):
                            await self._queueput(driver.readerque, xmldata)
                        elif devicename and propertyname and ((devicename, propertyname) in driver.snoopvectors):
                            await self._queueput(driver.readerque, xmldata)

                # now every driver/remcon which needs it has this xmldata
                # and the while loop now continues

        finally:
            self.shutdown()

    async def _sendtoclient(self):
        "For every clientconnection, get txque and copy data into it from serverwriterque"
        try:
            while not self._stop:
                try:
                    xmldata = await asyncio.wait_for(self.serverwriterque.get(), 0.5)
                except asyncio.TimeoutError:
                    continue
                self.serverwriterque.task_done()
                #  This xmldata of None is an indication to shut the server down
                #  It is set to None when a duplicate devicename is discovered
                if xmldata is None:
                    logger.error("A duplicate devicename has caused a server shutdown")
                    return
                if logger.isEnabledFor(logging.DEBUG) and self.debug_enable:
                    binarydata = ET.tostring(xmldata)
                    logger.debug(f"TX:: {binarydata.decode('utf-8')}")
                for clientconnection in self.connectionpool:
                    if clientconnection.connected:
                        await self._queueput(clientconnection.txque, xmldata)
                # The while loop continues
        finally:
            self.shutdown()



    async def send_message(self, message, timestamp=None):
        """Send system wide message, timestamp should normally not be set, if
           given, it should be a datetime.datetime object with tz set to timezone.utc"""
        if self._stop:
            return
        if not timestamp:
            timestamp = datetime.now(tz=timezone.utc).replace(tzinfo = None)
        else:
            if not isinstance(timestamp, datetime):
                # invalid timestamp given
                return
            if timestamp.tzinfo is not None:
                if timestamp.tzinfo == timezone.utc:
                    timestamp = timestamp.replace(tzinfo = None)
                else:
                    # invalid timestamp
                    return
        xmldata = ET.Element('message')
        xmldata.set("timestamp", timestamp.isoformat(sep='T'))
        xmldata.set("message", message)
        for clientconnection in self.connectionpool:
            if clientconnection.connected:
                # at least one is connected, so this data is put into
                # serverwriterque, and is then sent to each client by
                # the _sendtoclient method.
                await self._queueput(self.serverwriterque, xmldata)
                break

        for remcon in self.remotes:
            if not remcon.connected:
                continue
            await remcon.send(xmldata)


class _DriverComms:

    """An instance of this is created for each driver, which calls the __call__
       method.  Any data the driver wishes to be send will be taken
       from the drivers writerque and transmitted to the client by placing it
       into the serverwriterque"""

    def __init__(self, driver, serverwriterque, connectionpool, alldrivers, remotes):

        # This object is attached to this driver
        self.driver = driver
        self.serverwriterque = serverwriterque
        # connectionpool is a list of ClientConnection objects, which is used
        # to test if a client is connected
        self.connectionpool = connectionpool
        # self.connected is read by the driver, and in this case is always True
        # as the driver is connected to IPyServer, which handles snooping traffic,
        # even if no client is connected
        self.connected = True
        # self.alldrivers is set to a list of drivers, including exdrivers
        self.alldrivers = alldrivers
        # self.remotes is a list of connections to remote servers
        self.remotes = remotes
        self._stop = False       # Gets set to True to stop communications

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop

    def shutdown(self):
        "Sets self.stop to True and calls shutdown on tasks"
        self._stop = True

    async def _queueput(self, queue, value, timeout=0.5):
        while not self._stop:
            try:
                await asyncio.wait_for(queue.put(value), timeout)
            except asyncio.TimeoutError:
                # queue is full, continue while loop, checking stop flag
                continue
            break

    async def __call__(self, readerque, writerque):
        """Called by the driver, should run continuously.
           reads writerque from the driver, and sends xml data to the network"""
        while not self._stop:
            try:
                xmldata = await asyncio.wait_for(writerque.get(), 0.5)
            except asyncio.TimeoutError:
                continue
            writerque.task_done()
            # Check if other drivers/remotes wants to snoop this traffic
            devicename = xmldata.get("device")
            propertyname = xmldata.get("name")

            if xmldata.tag in NEWTAGS:
                # drivers should never transmit a new
                # but just in case
                logger.error(f"Driver transmitted invalid tag {xmldata.tag}")
                continue

            if xmldata.tag.startswith("def"):
                # check for duplicate devicename
                for driver in self.alldrivers:
                    if driver is self.driver:
                        continue
                    if devicename in driver:
                        logger.error(f"A duplicate devicename {devicename} has been detected")
                        await self._queueput(self.serverwriterque, None)
                        return
                for remote in self.remotes:
                    if devicename in remote.devicenames:
                        logger.error(f"A duplicate devicename {devicename} has been detected")
                        await self._queueput(self.serverwriterque, None)
                        return


            # check for a getProperties
            if xmldata.tag == "getProperties":
                foundflag = False
                # if getproperties is targetted at a known device, send it to that device
                if devicename:
                    for driver in self.alldrivers:
                        if driver is self.driver:
                            # No need to check sending a getProperties to itself
                            continue
                        if devicename in driver:
                            # this getProperties request is meant for an attached driver/device
                            await self._queueput(driver.readerque, xmldata)
                            foundflag = True
                            break
                    if foundflag:
                        # no need to transmit this anywhere else, continue the while loop
                        continue


            # transmit xmldata out to other drivers
            for driver in self.alldrivers:
                if driver is self.driver:
                    continue
                if xmldata.tag == "getProperties":
                    # either no devicename, or an unknown device
                    await self._queueput(driver.readerque, xmldata)
                else:
                    # Check if this driver is snooping on this device/vector
                    if driver.snoopall:
                        await self._queueput(driver.readerque, xmldata)
                    elif devicename and (devicename in driver.snoopdevices):
                        await self._queueput(driver.readerque, xmldata)
                    elif devicename and propertyname and ((devicename, propertyname) in driver.snoopvectors):
                        await self._queueput(driver.readerque, xmldata)


            for remcon in self.remotes:
                if not remcon.connected:
                    continue
                # send to all remotes
                await remcon.send(xmldata)

            for clientconnection in self.connectionpool:
                if clientconnection.connected:
                    # at least one is connected, so this data is put into
                    # serverwriterque, and is then sent to each client
                    await self._queueput(self.serverwriterque, xmldata)
                    break




class _ClientConnection:

    "Handles a client connection"

    def __init__(self, connection_id, serverreaderque):

        # number identifying this connection
        self.connection_id = connection_id

        # self.txque will have data to be transmitted
        # inserted into it from the IPyServer._sendtoclient()
        # method
        self.txque = asyncio.Queue(6)

        self.serverreaderque = serverreaderque
        # self.connected is True if this pool object is running a connection
        self.connected = False

        self.rx = None
        self.tx = None

        self._stop = False       # Gets set to True to stop communications

    @property
    def stop(self):
        "returns self._stop"
        return self._stop

    def shutdown(self):
        "Sets self.stop to True and calls shutdown on tasks"
        self._stop = True
        self.connected = False
        if self.rx is not None:
            self.rx.shutdown()
        if self.tx is not None:
            self.tx.shutdown()

    async def handle_data(self, reader, writer):
        "Used by asyncio.start_server, called to handle a client connection"
        self.connected = True
        sendchecker = SendChecker()
        addr = writer.get_extra_info('peername')
        self.rx = Conn_RX(self.connection_id, sendchecker, reader)
        self.tx = Conn_TX(sendchecker, writer)
        logger.info(f"Connection received from {addr}")
        try:
            txtask = asyncio.create_task(self.tx.run_tx(self.txque))
            rxtask = asyncio.create_task(self.rx.run_rx(self.serverreaderque))
            await asyncio.gather(txtask, rxtask)
        except ConnectionError:
            pass
        finally:
            self.connected = False
            txtask.cancel()
            rxtask.cancel()
        cleanque(self.txque)
        logger.info(f"Connection from {addr} closed")
        while True:
            if txtask.done() and rxtask.done():
                break
            await asyncio.sleep(1)



class Conn_TX():
    "An object that transmits data on a port"

    def __init__(self, sendchecker, writer):
        self.sendchecker = sendchecker
        self.writer = writer
        self._stop = False       # Gets set to True to stop communications

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop"
        return self._stop

    def shutdown(self):
        self._stop = True

    async def run_tx(self, writerque):
        """Gets data from writerque, and transmits it out on the port writer"""
        while not self._stop:
            await asyncio.sleep(0)
            # get block of data from writerque and transmit
            try:
                txdata = await asyncio.wait_for(writerque.get(), 0.5)
            except asyncio.TimeoutError:
                continue
            writerque.task_done()
            if txdata is None:
                continue
            if not self.sendchecker.allowed(txdata):
                # this data should not be transmitted, discard it
                continue
            # this data can be transmitted
            binarydata = ET.tostring(txdata)
            # Send to the port
            self.writer.write(binarydata)
            await self.writer.drain()
        self.writer.close()
        await self.writer.wait_closed()




class Conn_RX():
    """Produces xml.etree.ElementTree data from data received on the port"""

    def __init__(self, connection_id, sendchecker, reader):
        self._remainder = b""    # Used to store intermediate data
        self._stop = False       # Gets set to True to stop communications
        self.sendchecker = sendchecker
        self.reader = reader
        self.connection_id = connection_id

    def shutdown(self):
        self._stop = True

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop"
        return self._stop

    async def run_rx(self, serverreaderque):
        "pass xml.etree.ElementTree data to serverreaderque"
        try:
            # get block of xml.etree.ElementTree data
            # from self._xmlinput and append it to  serverreaderque together with connection_id
            while not self._stop:
                rxdata = await self._xmlinput()
                if rxdata is None:
                    return
                if rxdata.tag == "enableBLOB":
                    # set permission flags in the sendchecker object
                    self.sendchecker.setpermissions(rxdata)
                # and place rxdata into serverreaderque
                while not self._stop:
                    try:
                        await asyncio.wait_for(serverreaderque.put((self.connection_id, rxdata)), timeout=0.5)
                    except asyncio.TimeoutError:
                        # queue is full, continue while loop, checking stop flag
                        continue
                    # rxdata is now in serverreaderque, break the inner while loop
                    break
        except ConnectionError:
            # re-raise this without creating a report, as it probably indicates
            # a normal connection drop
            raise
        except Exception:
            # possibly some other error, so report it
            logger.exception("Exception report from Conn_RX.run_rx")
            raise


    async def _xmlinput(self):
        """get data from  _datainput, parse it, and return it as xml.etree.ElementTree object
           Returns None if stop flags arises"""
        message = b''
        messagetagnumber = None
        while not self._stop:
            await asyncio.sleep(0)
            data = await self._datainput()
            # data is either None, or binary data ending in b">"
            if data is None:
                return
            if self._stop:
                return
            if not message:
                # data is expected to start with <tag, first strip any newlines
                data = data.strip()
                for index, st in enumerate(_STARTTAGS):
                    if data.startswith(st):
                        messagetagnumber = index
                        break
                    elif st in data:
                        # remove any data prior to a starttag
                        positionofst = data.index(st)
                        data = data[positionofst:]
                        messagetagnumber = index
                        break
                else:
                    # data does not start with a recognised tag, so ignore it
                    # and continue waiting for a valid message start
                    continue
                # set this data into the received message
                message = data
                # either further children of this tag are coming, or maybe its a single tag ending in "/>"
                if message.endswith(b'/>'):
                    # the message is complete, handle message here
                    try:
                        root = ET.fromstring(message.decode("us-ascii"))
                    except Exception:
                        # failed to parse the message, continue at beginning
                        message = b''
                        messagetagnumber = None
                        continue
                    # xml datablock done, return it
                    return root
                # and read either the next message, or the children of this tag
                continue
            # To reach this point, the message is in progress, with a messagetagnumber set
            # keep adding the received data to message, until an endtag is reached
            message += data
            if message.endswith(_ENDTAGS[messagetagnumber]):
                # the message is complete, handle message here
                try:
                    root = ET.fromstring(message.decode("us-ascii"))
                except Exception:
                    # failed to parse the message, continue at beginning
                    message = b''
                    messagetagnumber = None
                    continue
                # xml datablock done, return it
                return root
            # so message is in progress, with a messagetagnumber set
            # but no valid endtag received yet, so continue the loop


    async def _datainput(self):
        """Waits for binary string of data ending in > from the port
           Returns None if stop flags arises"""
        binarydata = b""
        while not self._stop:
            await asyncio.sleep(0)
            try:
                data = await self.reader.readuntil(separator=b'>')
            except asyncio.LimitOverrunError:
                data = await self.reader.read(n=32000)
            except asyncio.IncompleteReadError:
                binarydata = b""
                await asyncio.sleep(0.1)
                continue
            if not data:
                await asyncio.sleep(0.1)
                continue
            # data received
            if b">" in data:
                binarydata = binarydata + data
                return binarydata
            # data has content but no > found
            binarydata += data
            # could put a max value here to stop this increasing indefinetly



# Command to control whether setBLOBs should be sent to this channel from a given Device. They can
# be turned off completely by setting Never (the default), allowed to be intermixed with other INDI
# commands by setting Also or made the only command by setting Only.

# <!ELEMENT enableBLOB %BLOBenable >
# <!ATTLIST enableBLOB
# device %nameValue; #REQUIRED  name of Device
# name %nameValue; #IMPLIED name of BLOB Property, or all if absent
# >



class SendChecker:
    """Carries the enableBLOB status on a connection, and does checks
       to ensure valid data is being transmitted"""

    def __init__(self):
        "For every device create a dictionary"
        self.devicestatus = {}
        # create a dictionary of devicenames : to devicedict
        # where the devicedict will be {"Default":"Never", "Properties":{}}
        # The Properties value is a dictionary of propertyname:status


    def allowed(self, xmldata):
        "Return True if this xmldata can be transmitted, False otherwise"

        if xmldata.tag in ("getProperties", "delProperty"):
            return True

        if xmldata.tag not in ("defBLOBVector", "setBLOBVector", 'newBLOBVector'):
            # so anything other than a BLOB
            if self.rxonly():
                # Only blobs allowed
                return False
            return True

        # so following checks only apply to BLOB vectors

        devicename = xmldata.get("device")

        if devicename not in self.devicestatus:
            # devicename not recognised, add it
            self.devicestatus[devicename] = {"Default":"Never", "Properties":{}}

        if xmldata.tag == "defBLOBVector":
            return True

        devicedict = self.devicestatus[devicename]

        # so we have a devicename, get propertyname
        name = xmldata.get("name")

        # so we have a devicename, property name,
        if name and (name in devicedict["Properties"]):
            if devicedict["Properties"][name] == "Never":
                return False
            else:
                return True
        elif devicedict["Default"] == "Never":
            return False
        else:
            return True

    def setpermissions(self, rxdata):
        "Read the received enableBLOB xml and set permission in self.devicestatus"
        devicename = rxdata.get("device")
        if devicename is None:
            # invalid
            return
        if devicename not in self.devicestatus:
            # devicename not recognised, add it
            self.devicestatus[devicename] = {"Default":"Never", "Properties":{}}

        # get the status of Never, Also, Only
        status = rxdata.text.strip()
        if status not in ("Never", "Also", "Only"):
            # invalid
            return

        devicedict = self.devicestatus[devicename]

        # property name
        name = rxdata.get("name")
        if name is None:
            # This applies to the device rather than to a particular property
            devicedict["Default"] = status
        else:
            # add it to devicedict, and hence to self.devicestatus
            devicedict["Properties"][name] = status

    def rxonly(self):
        "Returns True if any device or property has been set to BLOBs only"
        for devicedict in self.devicestatus.values():
            if devicedict["Default"] == "Only":
                return True
            properties = devicedict["Properties"]
            for status in properties.values():
                if status == "Only":
                    return True
        return False





import collections, asyncio, sys, time, os, fcntl, logging

import xml.etree.ElementTree as ET

from . import events
from .propertyvectors import timestamp_string
from .propertymembers import getfloat


logger = logging.getLogger(__name__)


# All xml data received from the client, or from snooped devices should be contained in one of the following tags
TAGS = (b'getProperties',
        b'newTextVector',
        b'newNumberVector',
        b'newSwitchVector',
        b'newBLOBVector',
        b'enableBLOB',
        # below are tags from snooped devices
        b'message',
        b'delProperty',
        b'defSwitchVector',
        b'setSwitchVector',
        b'defLightVector',
        b'setLightVector',
        b'defTextVector',
        b'setTextVector',
        b'defNumberVector',
        b'setNumberVector',
        b'defBLOBVector',
        b'setBLOBVector'
       )


# _STARTTAGS is a tuple of ( b'<defTextVector', ...  ) data received will be tested to start with such a starttag
_STARTTAGS = tuple(b'<' + tag for tag in TAGS)


# _ENDTAGS is a tuple of ( b'</defTextVector>', ...  ) data received will be tested to end with such an endtag
_ENDTAGS = tuple(b'</' + tag + b'>' for tag in TAGS)


def _makestart(element):
    "Given an xml element, returns a string of its start, including < tag attributes >"
    attriblist = ["<", element.tag]
    for key,value in element.attrib.items():
        attriblist.append(f" {key}=\"{value}\"")
    attriblist.append(">")
    return "".join(attriblist)


class IPyDriver(collections.UserDict):

    """A subclass of this should be created with methods written
       to control your device.

       devices are Device objects this driver handles.

       You may optionally include named arguments of any instrumentation
       objects that may be useful to you, these will be available in
       the dictionary self.driverdata.

       This object is also a mapping, of devicename:deviceobject
       """

    @staticmethod
    def indi_number_to_float(value):
        """The INDI spec allows a number of different number formats, given any number string, this returns a float.
           If an error occurs while parsing the number, a TypeError exception is raised."""
        return getfloat(value)


    def __init__(self, *devices, **driverdata):
        super().__init__()

        # self.data is defined as empty dictionary in UserDict
        for device in devices:
            devicename = device.devicename
            if devicename in self.data:
                # duplicate devicename
                raise ValueError(f"Device name {devicename} is duplicated in this driver.")
            self.data[devicename] = device
            # set driver into devices
            device.driver = self

        # self.data is used by UserDict, it will become
        # a dictionary of {devicename:device, ... }

        # dictionary of optional data
        self.driverdata = driverdata

        # traffic is transmitted out on the writerque
        self.writerque = asyncio.Queue(4)
        # and read in from the readerque
        self.readerque = asyncio.Queue(4)
        # and snoop data is passed on to the snoopque
        self.snoopque = asyncio.Queue(4)
        # data for each device is passed to each device dataque

        # An object for communicating can be set, if not set, then
        # self.comms = _STDINOUT() will be set in the asyncrun call
        self.comms = None

        # These set the remote traffic which this driver is snooping
        # initially the driver is not snooping anything, until it sends
        # a getProperties
        self.snoopall = False           # gets set to True if it is snooping everything
        self.snoopdevices = set()       # gets set to a set of device names

        self.snoopvectors = {}
        # The keys of self.snoopvectors will be tuples (devicename,vectorname)
        # of vectors that are to be snooped
        # The values will be either None or lists of [timeout, timestamp]

        # timeout is integer seconds set by the snoop() method
        # timestamp is updated whenever snoop data from devicename,vectorname
        # is received.
        # The coroutine _monitorsnoop Checks if current time is greater than
        # timeout+timestamp, and if it is, sends a getproperties

        self.debug_enable = False
        # If True, xmldata will be logged at DEBUG level

        self.auto_send_def = True
        # If True, whenever a getProperties event is received, a
        # vector send_defVector() will be called, automatically replying with
        # the vector definition.
        # If set to False, the driver developer will need to implement a send_defVector()
        # in the rxevent method

        self._stop = False
        # shutdown routine sets this to True to stop coroutines

        self.stopped = asyncio.Event()
        # this is set when asyncrun is finished


    def devices(self):
        "Returns a list of device objects"
        return list(self.data.values())

    def shutdown(self):
        "Shuts down the driver, sets the flag self.stop to True"
        self._stop = True
        if self.comms is not None:
            self.comms.shutdown()
        for device in self.data.values():
            device.shutdown()

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop


    async def _queueput(self, queue, value, timeout=0.5):
        while not self._stop:
            try:
                await asyncio.wait_for(queue.put(value), timeout)
            except asyncio.TimeoutError:
                # queue is full, continue while loop, checking stop flag
                continue
            break


    async def send(self, xmldata):
        "Transmits xmldata, this is an internal method, not normally called by a user."
        if not self.comms.connected:
            return
        while not self._stop:
            if not self.comms.connected:
                return
            try:
                await asyncio.wait_for(self.writerque.put(xmldata), timeout=0.5)
            except asyncio.TimeoutError:
                # queue is full, continue while loop, checking stop flag
                continue
            break
        if logger.isEnabledFor(logging.DEBUG) and self.debug_enable:
            binarydata = ET.tostring(xmldata)
            logger.debug(f"TX:: {binarydata.decode('utf-8')}")


    def __setitem__(self, devicename):
        raise KeyError

    async def _read_readerque(self):
        client_tags = ("enableBLOB", "newSwitchVector", "newNumberVector", "newTextVector", "newBLOBVector")
        snoop_tags = ("message", 'delProperty', 'defSwitchVector', 'setSwitchVector', 'defLightVector',
                      'setLightVector', 'defTextVector', 'setTextVector', 'defNumberVector', 'setNumberVector',
                      'defBLOBVector', 'setBLOBVector')
        while not self._stop:
            # reads readerque, and sends xml data to the device via its dataque
            try:
                root = await asyncio.wait_for(self.readerque.get(), 0.5)
            except asyncio.TimeoutError:
                continue
            self.readerque.task_done()
            # log the received data
            if logger.isEnabledFor(logging.DEBUG) and self.debug_enable:
                binarydata = ET.tostring(root)
                logger.debug(f"RX:: {binarydata.decode('utf-8')}")
            if root.tag == "getProperties":
                version = root.get("version")
                if version != "1.7":
                    continue
                # getProperties received with correct version
                devicename = root.get("device")
                # devicename is None (for all devices), or a named device
                if devicename is None:
                    for d in self.data.values():
                        if d.enable:
                            await self._queueput(d.dataque, root)
                elif devicename in self.data:
                    if self.data[devicename].enable:
                        await self._queueput(self.data[devicename].dataque, root)
                # else device not recognised
            elif root.tag in client_tags:
                # xml received from client
                devicename = root.get("device")
                if devicename is None:
                    # device not given, ignore this
                    continue
                elif devicename in self.data:
                    if self.data[devicename].enable:
                        await self._queueput(self.data[devicename].dataque, root)
                # else device not recognised
            elif root.tag in snoop_tags:
                # xml received from other devices
                await self._queueput(self.snoopque, root)


    async def _call_snoopevent(self, event):
        "Update timestamp when snoop data received and call self.snoopevent"
        if event.devicename and event.vectorname:
            # update timestamp in self.snoopvectors
            timedata = self.snoopvectors.get((event.devicename, event.vectorname))
            if timedata is not None:
                timedata[1] = time.time()
        await self.snoopevent(event)


    async def _snoophandler(self):
        """Creates events using data from self.snoopque"""
        while not self._stop:
            # get block of data from the self.snoopque
            try:
                root = await asyncio.wait_for(self.snoopque.get(), 0.5)
            except asyncio.TimeoutError:
                continue
            self.snoopque.task_done()
            devicename = root.get("device")
            if devicename is not None:
                # if a device name is given, check
                # it is not in this drivers devices
                if devicename in self.data:
                    logger.error("Cannot snoop on a device already controlled by this driver")
                    continue
            try:
                if root.tag == "message":
                    # create event
                    event = events.message(root)
                    await self._call_snoopevent(event)
                elif root.tag == "delProperty":
                    # create event
                    event = events.delProperty(root)
                    await self._call_snoopevent(event)
                elif root.tag == "defSwitchVector":
                    # create event
                    event = events.defSwitchVector(root)
                    await self._call_snoopevent(event)
                elif root.tag == "setSwitchVector":
                    # create event
                    event = events.setSwitchVector(root)
                    await self._call_snoopevent(event)
                elif root.tag == "defLightVector":
                    # create event
                    event = events.defLightVector(root)
                    await self._call_snoopevent(event)
                elif root.tag == "setLightVector":
                    # create event
                    event = events.setLightVector(root)
                    await self._call_snoopevent(event)
                elif root.tag == "defTextVector":
                    # create event
                    event = events.defTextVector(root)
                    await self._call_snoopevent(event)
                elif root.tag == "setTextVector":
                    # create event
                    event = events.setTextVector(root)
                    await self._call_snoopevent(event)
                elif root.tag == "defNumberVector":
                    # create event
                    event = events.defNumberVector(root)
                    await self._call_snoopevent(event)
                elif root.tag == "setNumberVector":
                    # create event
                    event = events.setNumberVector(root)
                    await self._call_snoopevent(event)
                elif root.tag == "defBLOBVector":
                    # create event
                    event = events.defBLOBVector(root)
                    await self._call_snoopevent(event)
                elif root.tag == "setBLOBVector":
                    # create event
                    event = events.setBLOBVector(root)
                    await self._call_snoopevent(event)
            except events.EventException:
                # if an EventException is raised, it is because received data is malformed
                # so log it
                logger.exception("An exception occurred creating a snoop event")


    async def send_message(self, message="", timestamp=None):
        "Send system wide message - without device name"
        tstring = timestamp_string(timestamp)
        if not tstring:
            logger.error("The timestamp given in send_message must be a datetime.datetime UTC object")
            return
        xmldata = ET.Element('message')
        xmldata.set("timestamp", tstring)
        if message:
            xmldata.set("message", message)
        await self.send(xmldata)


    def snoop(self, devicename, vectorname, timeout=30):
        """Call this to snoop on a given devicename, vectorname.
           This will cause a getProperties to be sent, and will also
           send further getProperties every timeout seconds if no snooping
           data is being received from the specified vector.
           This avoids a possible problem where intermediate servers may
           be temporarily turned off, and will lose their instruction to
           broadcast snooping traffic. This method is only applicable when
           snooping on a specific device vector.
           timeout must be an integer equal or greater than 5 seconds."""
        if devicename in self.data:
            logger.error("Cannot snoop on a device already controlled by this driver")
            return
        timeout = int(timeout)
        if timeout < 5:
            logger.error("Snoop timout should be equal or greater than 5 seconds")
            return

        current = time.time()

        # set self.snoopvectors[(devicename,vectorname)] to [timeout, timestamp]

        self.snoopvectors[(devicename,vectorname)] = [timeout, current - timeout + 1]

        # setting timestamp to current - timeout + 1 means that after a second
        # the coroutine _monitorsnoop will think that its own time measurement
        # is greater than the timestamp plus timeout and will send a send_getProperties


    async def _monitorsnoop(self):
        "Checks if any snooping vectors have timed out, if it has, sends getproperties"
        while not self._stop:
            await asyncio.sleep(1)
            if not self.snoopvectors:
                continue
            current = time.time()
            for key, value in self.snoopvectors.items():
                if value is None:
                    continue
                timeout, timestamp = value
                if current > timestamp + timeout:
                    # the timeout has expired, update timestamp and send getproperties
                    value[1] = current
                    await self.send_getProperties(*key)


    async def send_getProperties(self, devicename=None, vectorname=None):
        """Sends a getProperties request - which is used to snoop data from other devices
           on the network, if devicename given, it must not be a device of this driver as
           the point of this is to snoop on remote devices."""
        xmldata = ET.Element('getProperties')
        xmldata.set("version", "1.7")
        if devicename is None:
            await self.send(xmldata)
            self.snoopall = True
            return
        if devicename in self.data:
            logger.error("Cannot snoop on a device already controlled by this driver")
            return
        xmldata.set("device", devicename)
        if vectorname is None:
            await self.send(xmldata)
            self.snoopdevices.add(devicename)
            return
        xmldata.set("name", vectorname)
        await self.send(xmldata)
        # adds tuple (devicename,vectorname) to self.snoopvectors
        if (devicename,vectorname) not in self.snoopvectors:
            self.snoopvectors[(devicename,vectorname)] = None


    async def hardware(self):
        """Override this to operate device hardware, and transmit updates

        For example: call your own code to operate hardware
        then update the appropriate vectors, and send updated
        values to the client using
        await vector.send_setVector()"""
        pass


    async def rxevent(self, event):
        """Override this. On receiving data, this is called, and should
           handle any necessary actions.
           event is an object describing the event, with attributes
           devicename, vectorname, vector, root
           where vector is the properties vector the event refers to, and
           root is an xml.etree.ElementTree object of the received xml"""
        pass


    async def snoopevent(self, event):
        """Override this if this driver is snooping on other devices.
           On receiving snoop data, this is called, and should handle
           any necessary actions.
           event is an object with attributes according to the data received."""
        pass


    async def rxgetproperties(self, event):
        """This is an internal method called whenever a getProperties
           event is received from the client. It replies with a send_defVector
           to send a property vector definition back to the client.
           It would normally never be called by users own code.
           If the auto_send_def attribute is False, this does not send anything,
           and the getProperties event needs to be handled by rxevent."""
        if self.auto_send_def:
            await event.vector.send_defVector()
        else:
            await self.rxevent(event)

    async def asyncrun(self):
        """await this to operate the driver, which will then communicate by
           stdin and stdout.

           Do not await this if the driver is being set into IPyServer, in
           that situation the IPyServer will control communications."""

        logger.info(f"Driver {self.__class__.__name__} started")
        self._stop = False

        # set an object for communicating, as default this is stdin and stdout
        if self.comms is None:
            self.comms = _STDINOUT()

        # get all tasks into a list

        tasks = [ self.comms(self.readerque, self.writerque),    # run communications
                  self.hardware(),                               # task to operate device hardware, and transmit updates
                  self._read_readerque(),                        # task to handle received xml data
                  self._monitorsnoop(),                          # task to monitor if a getproperties needs to be sent
                  self._snoophandler() ]                         # task to handle incoming snoop data

        for device in self.data.values():
            tasks.append(device._handler())                           # each device handles its incoming data
            for pv in device.values():
                tasks.append(pv._handler())                           # each property handles its incoming data
                # also give the propertyvector a reference to this driver
                # so it can call eventaction and have access to writerque
                pv.driver = self
        try:
            await asyncio.gather( *tasks )
        finally:
            self.stopped.set()
            self._stop = True


class Device(collections.UserDict):

    """An instance of this should be created for each device controlled by this driver.
       The properties argument is a list of vectors controlling this device.
       devicedata will be an attribute dictionary of any hardware data that may be usefull.

       This object will be a mapping of vector name to vector object
    """

    def __init__(self, devicename, properties, **devicedata):
        super().__init__()

        if not devicename.isascii():
            raise ValueError("Device name should be all ascii characters.")

        # This device name
        self.devicename = devicename

        # if self.enable is False, this device ignores incoming traffic
        # and (apart from delProperty) does not transmit anything
        # from his device
        self.enable = True

        # the driver places data in this que to send data to this device
        self.dataque = asyncio.Queue(4)

        # Every property of this device has a dataque, which is set into this dictionary
        self.propertyquedict = {p.name:p.dataque for p in properties}

        # dictionary of optional data
        self.devicedata = devicedata

        # this will be set when the driver asyncrun is run
        self.driver = None

        # self.data is a dictionary of name to vector this device owns
        for p in properties:
            p.devicename = self.devicename
            self.data[p.name] = p

        # shutdown routine sets this to True to stop coroutines
        self._stop = False


    def properties(self):
        "Returns a list of vector objects"
        return list(self.data.values())


    def shutdown(self):
        """Shuts down the device, sets the flag self._stop to True
           and shuts down property vector handlers"""
        self._stop = True
        for pv in self.data.values():
            pv.shutdown()

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop


    async def _queueput(self, queue, value, timeout=0.5):
        while not self._stop:
            try:
                await asyncio.wait_for(queue.put(value), timeout)
            except asyncio.TimeoutError:
                # queue is full, continue while loop, checking stop flag
                continue
            break

    async def send_device_message(self, message="", timestamp=None):
        """Send a message associated with this device, which the client could display.
           The timestamp should be either None or a datetime.datetime object. If the
           timestamp is None a UTC value will be inserted."""
        if not self.enable:
            # messages only sent if self.enable is True
            return
        tstring = timestamp_string(timestamp)
        if not tstring:
            logger.error("The timestamp given in send_device_message must be a datetime.datetime UTC object")
            return
        xmldata = ET.Element('message')
        xmldata.set("device", self.devicename)
        xmldata.set("timestamp", tstring)
        if message:
            xmldata.set("message", message)
        await self.driver.send(xmldata)

    async def send_delProperty(self, message="", timestamp=None):
        """Sending delProperty with this device method, (as opposed to the vector send_delProperty method)
           informs the client this device is not available, it also sets a device.enable attribute to
           False, which stops any data being transmitted between the client and this device.
           Setting device.enable to True re-enables communications.
           The message argument is any appropriate string which the client could display to the user.
           The timestamp should be either None or a datetime.datetime object. If the timestamp is None
           a UTC value will be inserted."""
        tstring = timestamp_string(timestamp)
        if not tstring:
            logger.error("The timestamp given in send_delProperty must be a datetime.datetime UTC object")
            return
        xmldata = ET.Element('delProperty')
        xmldata.set("device", self.devicename)
        xmldata.set("timestamp", tstring)
        if message:
            xmldata.set("message", message)
        await self.driver.send(xmldata)
        self.enable = False


    async def devhardware(self, *args, **kwargs):
        """As default, does nothing and is not called.

           If required, override this to handle any necessary device actions.
           This should be called from the driver 'hardware' method if it is used."""
        pass


    async def devrxevent(self, event, *args, **kwargs):
        """As default, does nothing and is not called.

           If required, override this to handle any necessary device actions.
           This should be called from the driver 'rxevent' method if it is used."""
        pass


    async def devsnoopevent(self, event, *args, **kwargs):
        """As default, does nothing and is not called.

           If required, override this to handle any necessary device actions.
           This should be called from the driver 'snoopevent' method if it is used."""
        pass


    def __setitem__(self, vectorname):
        raise KeyError

    async def _handler(self):
        """Handles data read from dataque"""
        while not self._stop:
            # get block of data from the self.dataque
            try:
                root = await asyncio.wait_for(self.dataque.get(), 0.5)
            except asyncio.TimeoutError:
                continue
            self.dataque.task_done()
            if not self.enable:
                continue
            if root.tag == "getProperties":
                name = root.get("name")
                # name is None (for all properties), or a named property
                if name is None:
                    for pvector in self.data.values():
                        if pvector.enable:
                            await self._queueput(pvector.dataque, root)
                elif name in self.data:
                    if self.data[name].enable:
                        await self._queueput(self.data[name].dataque, root)
                else:
                    # property name not recognised
                    continue
            elif root.tag == "enableBLOB":
                name = root.get("name")
                # name is None (for all properties), or a named property
                if name is None:
                    for pvector in self.data.values():
                        if pvector.enable:
                            await self._queueput(pvector.dataque, root)
                elif name in self.data:
                    if self.data[name].enable:
                        await self._queueput(self.data[name].dataque, root)
                else:
                    # property name not recognised
                    continue
            else:
                # root.tag will be one of
                # newSwitchVector, newNumberVector, newTextVector, newBLOBVector
                name = root.get("name")
                if name is None:
                    # name not given, ignore this
                    continue
                elif name in self.data:
                    pvector = self.data[name]
                    if pvector.perm != "ro" and pvector.enable:
                        # all ok, add to the vector dataque
                        await self._queueput(pvector.dataque, root)


class _STDOUT_TX:
    "An object that transmits data on stdout, used by _STDINOUT as one half of the communications path"

    def __init__(self):
        self._stop = False       # Gets set to True to stop communications

    def shutdown(self):
        self._stop = True

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop


    async def run_tx(self, writerque):
        """Gets data from writerque, and transmits it out on stdout"""
        while not self._stop:
            await asyncio.sleep(0)
            # get block of data from writerque and transmit down stdout
            try:
                txdata = await asyncio.wait_for(writerque.get(), 0.5)
            except asyncio.TimeoutError:
                # test self._stop again
                continue
            writerque.task_done()
            if txdata is None:
                await asyncio.sleep(0.02)
                continue
            if (txdata.tag == "setBLOBVector") and len(txdata):
                # txdata is a setBLOBVector containing blobs
                # send initial setBLOBVector
                startdata = _makestart(txdata)
                sys.stdout.buffer.write(startdata.encode())
                sys.stdout.buffer.flush()
                for oneblob in txdata.iter('oneBLOB'):
                    bytescontent = oneblob.text.encode()
                    # send start of oneblob
                    startoneblob = _makestart(oneblob)
                    sys.stdout.buffer.write(startoneblob.encode())
                    sys.stdout.buffer.flush()
                    # send content in chunks
                    chunksize = 1000
                    for b in range(0, len(bytescontent), chunksize):
                        byteschunk = bytescontent[b:b+chunksize]
                        sys.stdout.buffer.write(byteschunk)
                        sys.stdout.buffer.flush()
                        await asyncio.sleep(0)
                    sys.stdout.buffer.write(b"</oneBLOB>")
                    sys.stdout.buffer.flush()
                # send enddata
                sys.stdout.buffer.write(b"</setBLOBVector>\n")
                sys.stdout.buffer.flush()
            else:
                # its straight xml, send it out on stdout
                binarydata = ET.tostring(txdata)
                binarydata += b"\n"
                sys.stdout.buffer.write(binarydata)
                sys.stdout.buffer.flush()


class _STDIN_RX:
    """An object that receives data on stdin, parses it to ElementTree elements
       and passes it to the driver by appending it to the driver's readerque"""

    def __init__(self):
        self._remainder = b""    # Used to store intermediate data
        self._stop = False       # Gets set to True to stop communications

    def shutdown(self):
        self._stop = True

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop

    async def run_rx(self, readerque):
        "pass data to readerque"
        try:
            # get block of xml.etree.ElementTree data
            # from self._xmlinput and append it to readerque
            while not self._stop:
                rxdata = await self._xmlinput()
                if rxdata is None:
                    return
                # append it to readerque
                while not self._stop:
                    try:
                        await asyncio.wait_for(readerque.put(rxdata), timeout=0.5)
                    except asyncio.TimeoutError:
                        # queue is full, continue while loop, checking stop flag
                        continue
                    # rxdata is now in readerque, break the inner while loop
                    break
        except Exception:
            logger.exception("Exception report from _STDIN_RX.run_rx")
            raise

    async def _xmlinput(self):
        """get data from  _datainput, parse it, and return it as xml.etree.ElementTree object
           Returns None if stop flags arises"""
        message = b''
        messagetagnumber = None
        while not self._stop:
            await asyncio.sleep(0)
            data = await self._datainput()
            # data is either None, or binary data ending in b">"
            if data is None:
                return
            if self._stop:
                return
            if not message:
                # data is expected to start with <tag, first strip any newlines
                data = data.strip()
                for index, st in enumerate(_STARTTAGS):
                    if data.startswith(st):
                        messagetagnumber = index
                        break
                    elif st in data:
                        # remove any data prior to a starttag
                        positionofst = data.index(st)
                        data = data[positionofst:]
                        messagetagnumber = index
                        break
                else:
                    # data does not start with a recognised tag, so ignore it
                    # and continue waiting for a valid message start
                    continue
                # set this data into the received message
                message = data
                # either further children of this tag are coming, or maybe its a single tag ending in "/>"
                if message.endswith(b'/>'):
                    # the message is complete, handle message here
                    try:
                        root = ET.fromstring(message.decode("us-ascii"))
                    except Exception:
                        # failed to parse the message, continue at beginning
                        message = b''
                        messagetagnumber = None
                        continue
                    # xml datablock done, return it
                    return root
                # and read either the next message, or the children of this tag
                continue
            # To reach this point, the message is in progress, with a messagetagnumber set
            # keep adding the received data to message, until an endtag is reached
            message += data
            if message.endswith(_ENDTAGS[messagetagnumber]):
                # the message is complete, handle message here
                try:
                    root = ET.fromstring(message.decode("us-ascii"))
                except Exception:
                    # failed to parse the message, continue at beginning
                    message = b''
                    messagetagnumber = None
                    continue
                # xml datablock done, return it
                return root
            # so message is in progress, with a messagetagnumber set
            # but no valid endtag received yet, so continue the loop


    async def _datainput(self):
        """Waits for binary string of data ending in > from stdin
           Returns None if stop flags arises"""
        remainder = self._remainder
        if b">" in remainder:
            # This returns with binary data ending in > as long
            # as there are > characters in self._remainder
            binarydata, self._remainder = remainder.split(b'>', maxsplit=1)
            binarydata += b">"
            return binarydata
        # As soon as there are no > characters left in self._remainder
        # get more data from stdin
        while not self._stop:
            await asyncio.sleep(0)
            indata = sys.stdin.buffer.read(100)
            if not indata:
                await asyncio.sleep(0.02)
                continue
            remainder += indata
            if b">" in indata:
                binarydata, self._remainder = remainder.split(b'>', maxsplit=1)
                binarydata += b">"
                return binarydata



class _STDINOUT():
    """If indipydriver.comms is set to an instance of this class it is
       used to implement communications via stdin and stdout"""

    def __init__(self):
        self.connected = True
        self.rx = _STDIN_RX()
        self.tx = _STDOUT_TX()
        self._stop = False       # Gets set to True to stop communications


    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop

    def shutdown(self):
        self._stop = True
        self.rx.shutdown()
        self.tx.shutdown()

    async def __call__(self, readerque, writerque):
        "Called from indipydriver.asyncrun() to run the communications"
        # Set stdin to non-blocking mode
        flags = fcntl.fcntl(sys.stdin.fileno(), fcntl.F_GETFL)
        fcntl.fcntl(sys.stdin.fileno(), fcntl.F_SETFL, flags | os.O_NONBLOCK)

        logger.info("Communicating via STDIN/STDOUT")

        await asyncio.gather(self.rx.run_rx(readerque),
                             self.tx.run_tx(writerque)
                             )
